# -*- coding: utf-8 -*-
"""Utilities for chunking time-series data before feeding it to the operators.
"""

__author__ = 'Jonas Van Der Donckt'

from .chunking import chunk_data

__all__ = [
    "chunk_data"
]
